﻿#include "mirage.h"

#include "async/thread_pool.h"
#include "misc/lazy_singleton.h"
#include "renderer/dpi_sensibility.h"
#include "window/window_manager.h"

mirage::duration_type delta_time  = {};
mirage::time_type begin_time      = {};
mirage::time_type last_time       = {};
LLGL::RenderSystemPtr renderer    = nullptr;
uint64_t frame_index              = 0;
std::filesystem::path shader_path = "resource/shaders";

namespace mirage {
    void on_llgl_log(LLGL::Log::ReportType type, const char* text, void* user_data) {
        switch (type) {
        case LLGL::Log::ReportType::Default:
            spdlog::info(text);
            break;
        case LLGL::Log::ReportType::Error:
            spdlog::error(text);
            break;
        }
    }

    std::string to_string(renderer_api api) {
        switch (api) {
        case renderer_api::dx11:
            return "Direct3D11";
        case renderer_api::dx12:
            return "Direct3D12";
        case renderer_api::vulkan:
            return "Vulkan";
        case renderer_api::opengl:
            return "OpenGL";
        case renderer_api::metal:
            return "Metal";
        }
        return "Unknown";
    }

    bool init_renderer(const init_info& in_info) {
        dpi_sensibility::init_dpi_sensibility();
        RegisterCallback(on_llgl_log);
        LLGL::RenderSystemDescriptor desc{};
        desc.moduleName = to_string(in_info.api);

        spdlog::info("初始化渲染器: {}", desc.moduleName.c_str());
        renderer = LLGL::RenderSystem::Load(desc);
        if (!renderer) {
            spdlog::error("渲染器加载失败");
            return false;
        }
        auto renderer_info = renderer->GetRendererInfo();
        spdlog::info("渲染器信息: {}", renderer_info.rendererName.c_str());
        spdlog::info("使用设备: {}", renderer_info.deviceName.c_str());
        spdlog::info("供应商: {}", renderer_info.vendorName.c_str());
        spdlog::info("着色器语言: {}", renderer_info.shadingLanguageName.c_str());
        spdlog::info("渲染器加载成功");

        return true;
    }
    bool init_window(const init_info& in_info) {
        if (!lazy_singleton<window_manager>::get().init(in_info)) {
            spdlog::error("窗口管理器初始化失败");
            return false;
        }
        spdlog::info("窗口管理器创建成功");
        return true;
    }
    void destroy_renderer() {
        LLGL::RenderSystem::Unload(std::move(renderer));
    }

    bool init(const init_info& in_info) {
        spdlog::info("初始化 mirage");
        if (!init_renderer(in_info)) {
            return false;
        }
        if (!init_window(in_info)) {
            return false;
        }
        begin_time = std::chrono::high_resolution_clock::now();
        last_time  = std::chrono::high_resolution_clock::now();
        return true;
    }

    void destroy() {
        if (!lazy_singleton<window_manager>::get().destroy()) {
            spdlog::warn("窗口管理器销毋失败");
        }
        destroy_renderer();
        spdlog::info("mirage 销毁");
    }

    void update() {
        thread_pool::global().process_main_thread_callbacks();

        // 更新时间
        const auto& current_time = std::chrono::high_resolution_clock::now();
        delta_time               = current_time - last_time;
        last_time                = current_time;

        lazy_singleton<window_manager>::get().update(delta_time);
        frame_index++;

        std::this_thread::yield();
    }

    bool should_exit() {
        return lazy_singleton<window_manager>::get().should_exit();
    }

    renderer_api get_default_renderer_api() {
#if MIRAGE_PLATFORM_WINDOWS
        return renderer_api::dx11;
#elif MIRAGE_PLATFORM_MACOS
        return renderer_api::metal;
#elif MIRAGE_PLATFORM_LINUX
        return renderer_api::vulkan;
#endif
    }

    int run(const init_info& in_init_info) {
        try {
            if (!init(in_init_info)) {
                spdlog::error("初始化 mirage 失败");
                return -1;
            }
            while (!should_exit()) {
                update();
            }
            destroy();
        } catch (const std::exception& e) {
            spdlog::error("运行时异常: {}", e.what());
            return -1;
        } catch (...) {
            spdlog::error("未知异常");
            return -1;
        }
        return 0;
    }

    const duration_type& get_delta_time() {
        return delta_time;
    }

    duration_type get_total_time() {
        return std::chrono::high_resolution_clock::now() - begin_time;
    }
    LLGL::RenderSystem* get_renderer() {
        return renderer.get();
    }
    LLGL::CommandQueue* get_main_command_queue() {
        return renderer->GetCommandQueue();
    }
    uint64_t get_frame_index() {
        return frame_index;
    }

    std::filesystem::path get_shader_path() {
        return shader_path;
    }

    void set_shader_path(const std::filesystem::path& in_shader_path) {
        shader_path = in_shader_path;
    }
} // namespace mirage
