#pragma once
#include <chrono>
#include "LLGL/LLGL.h"
#include <Eigen/Eigen>
#include "LLGL/Utils/VertexFormat.h"

namespace mirage {
    using time_type = decltype(std::chrono::high_resolution_clock::now());
    using duration_type = decltype(std::chrono::high_resolution_clock::now() - std::chrono::high_resolution_clock::now());
    using pipeline_state_ptr = std::shared_ptr<LLGL::PipelineState>;
    using shader_ptr = std::shared_ptr<LLGL::Shader>;
    using pipeline_layout_ptr = std::shared_ptr<LLGL::PipelineLayout>;

    enum class renderer_api {
        dx11,
        dx12,
        vulkan,
        opengl,
        metal,
    };

    struct window_descriptor {
        std::u8string title;
        LLGL::Offset2D position;
        LLGL::Extent2D size;
        bool visible = true;
        bool borderless = false;
        bool resizable = true;
        bool centered  = false;
        bool accept_drop_files = false;
        bool disable_clear_on_resize = false;
        bool disable_size_scaling = false;

        operator LLGL::WindowDescriptor() const {
            LLGL::WindowDescriptor desc{};
            desc.title    = reinterpret_cast<const char*>(title.c_str());
            desc.position = position;
            desc.size     = size;
            desc.flags    = 0;
            if (visible) {
                desc.flags |= LLGL::WindowFlags::Visible;
            }
            if (borderless) {
                desc.flags |= LLGL::WindowFlags::Borderless;
            }
            if (resizable) {
                desc.flags |= LLGL::WindowFlags::Resizable;
            }
            if (centered) {
                desc.flags |= LLGL::WindowFlags::Centered;
            }
            if (accept_drop_files) {
                desc.flags |= LLGL::WindowFlags::AcceptDropFiles;
            }
            if (disable_clear_on_resize) {
                desc.flags |= LLGL::WindowFlags::DisableClearOnResize;
            }
            if (disable_size_scaling) {
                desc.flags |= LLGL::WindowFlags::DisableSizeScaling;
            }
            return desc;
        }
    };

    inline LLGL::VertexFormat create_vertex_format() {
        LLGL::VertexFormat vertex_format;
        vertex_format.AppendAttribute({"POSITION", 0, LLGL::Format::RG32Float});   // float2 position
        vertex_format.AppendAttribute({"TEXCOORD", 0, LLGL::Format::RG32Float});   // float2 uv
        vertex_format.AppendAttribute({"COLOR", 0, LLGL::Format::RGBA32Float});    // float4 color
        vertex_format.AppendAttribute({"TEXCOORD", 1, LLGL::Format::RGBA32Float}); // float4 param_a
        vertex_format.AppendAttribute({"TEXCOORD", 2, LLGL::Format::RGBA32Float}); // float4 param_b
        vertex_format.AppendAttribute({"TEXCOORD", 3, LLGL::Format::RGBA32Float}); // float4 param_c
        return vertex_format;
    }

    struct swap_chain_descriptor {
        int color_bits = 32;
        int depth_bits = 0;
        int stencil_bits = 0;
        std::uint32_t samples = 1;
        std::uint32_t swap_buffers = 2;
        bool fullscreen = false;
        bool vsync = true;

        operator LLGL::SwapChainDescriptor() const {
            LLGL::SwapChainDescriptor desc{};
            desc.colorBits = color_bits;
            desc.depthBits = depth_bits;
            desc.stencilBits = stencil_bits;
            desc.samples = samples;
            desc.swapBuffers = swap_buffers;
            desc.fullscreen = fullscreen;
            return desc;
        }
    };

    struct triangle_index_t {
        uint32_t vertex_index[3];
    };

    struct pipeline_info {
        pipeline_state_ptr pipeline_state;
        pipeline_layout_ptr pipeline_layout;
        std::vector<shader_ptr> shaders;
    };
}
