#include "render_buffer.h"

#include "render_command_pool.h"

namespace mirage {
    render_buffer::~render_buffer() {
        if (buffer) {
            get_renderer()->Release(*buffer);
        }
    }
    void render_buffer::init(const LLGL::BufferDescriptor& in_desc) {
        if (buffer) {
            throw std::runtime_error("Buffer already initialized");
        }

        LLGL::BufferDescriptor desc = in_desc;

        buffer = get_renderer()->CreateBuffer(desc);
        if (!buffer) {
            throw std::runtime_error("Failed to create buffer");
        }

        capacity     = desc.size;
        access_flags = desc.cpuAccessFlags;
    }
    void render_buffer::reserve(size_t new_capacity) {
        if (new_capacity > capacity) {
            resize(new_capacity, true);
        }
    }
    void render_buffer::shrink_to_fit() {
        if (capacity > size) {
            resize(size, true);
        }
    }
    void render_buffer::resize(size_t in_capacity_size, bool in_keep_data) {
        if (in_capacity_size <= capacity) {
            return;
        }
        LLGL::BufferDescriptor desc = buffer->GetDesc();
        desc.size                   = in_capacity_size;
        auto new_buffer             = get_renderer()->CreateBuffer(desc);
        if (!new_buffer) {
            spdlog::error("无法创建缓冲区");
            return;
        }

        if (in_keep_data) {
            const auto old_data = read();
            get_renderer()->WriteBuffer(*new_buffer, 0, old_data->get_data(), old_data->get_size_bytes());
        } else {
            size = 0;
        }

        if (buffer) {
            get_renderer()->Release(*buffer);
        }
        buffer   = new_buffer;
        capacity = in_capacity_size;
    }
    void render_buffer::clear() {
        size = 0;
    }
    heap_ptr render_buffer::read(size_t in_offset, size_t in_size) const {
        if (in_offset + in_size > size) {
            spdlog::error("读取缓冲区越界");
            return {};
        }
        auto out_heap = std::make_shared<heap>(in_size);
        get_renderer()->ReadBuffer(*buffer, in_offset, *out_heap, in_size);
        return out_heap;
    }
    heap_ptr render_buffer::read() const {
        return read(0, size);
    }
    void render_buffer::push(const void* data, size_t in_data_size) {
        if (size + in_data_size > capacity) {
            // 使用更智能的增长策略
            const size_t new_capacity = std::max(static_cast<size_t>(capacity * GROWTH_FACTOR), size + in_data_size);
            resize(new_capacity);
        }

        // 批量写入优化
        if (access_flags & LLGL::CPUAccessFlags::Write) {
            // 如果支持直接写入，使用映射方式
            if (auto mapped_data = get_renderer()->MapBuffer(*buffer, LLGL::CPUAccess::WriteOnly)) {
                std::memcpy(static_cast<uint8_t*>(mapped_data) + size, data, in_data_size);
                get_renderer()->UnmapBuffer(*buffer);
            }
        } else {
            // 否则使用WriteBuffer
            get_renderer()->WriteBuffer(*buffer, size, data, in_data_size);
        }
        size += in_data_size;
    }

    void render_buffer::update(const void* data, size_t offset, size_t size) {
        if (offset + size > capacity) {
            throw std::out_of_range("Buffer update out of bounds");
        }
        // 更新缓冲区
        get_renderer()->WriteBuffer(*buffer, offset, data, size);
    }

    void render_buffer::set(const void* in_data, size_t in_size) {
        if (in_size > capacity) {
            resize(in_size);
        }
        if (access_flags & LLGL::CPUAccessFlags::Write) {
            if (auto mapped = get_renderer()->MapBuffer(*buffer, LLGL::CPUAccess::WriteDiscard)) {
                std::memcpy(mapped, in_data, in_size);
                get_renderer()->UnmapBuffer(*buffer);
            }
        } else {
            get_renderer()->WriteBuffer(*buffer, 0, in_data, in_size);
        }
        size = in_size;
    }

    void render_buffer::update_async(const void* data, size_t offset, size_t size) {
        if (offset + size > capacity) {
            throw std::out_of_range("Buffer update out of bounds");
        }

        if (offset + size > capacity) {
            throw std::out_of_range("Buffer update out of bounds");
        }

        // 使用 RAII 方式获取命令缓冲区
        const scoped_command cmd;
        if (!cmd.is_valid()) {
            throw std::runtime_error("Failed to acquire command buffer");
        }

        // 复制数据到临时缓冲区（如果需要）
        std::vector temp_data(static_cast<const uint8_t*>(data), static_cast<const uint8_t*>(data) + size);

        // 更新缓冲区
        cmd->UpdateBuffer(*buffer, offset, temp_data.data(), size);

        // 提交到命令队列（scoped_command 析构时会自动 End）
        get_main_command_queue()->Submit(*cmd);
    }
    void render_buffer::batch_update_async(const std::vector<update_info>& updates) {
        scoped_command cmd;
        if (!cmd.is_valid()) {
            throw std::runtime_error("Failed to acquire command buffer");
        }

        for (const auto& update : updates) {
            if (update.offset + update.size > capacity) {
                spdlog::warn("跳过越界更新");
                continue;
            }
            cmd->UpdateBuffer(*buffer, update.offset, update.data, update.size);
        }

        get_main_command_queue()->Submit(*cmd);
    }
} // namespace mirage
