#pragma once
#include "LLGL/LLGL.h"
#include "mirage.h"
#include "rendering_common.h"
#include <span>

namespace mirage {
    struct heap {
        explicit heap(size_t in_size) : size(in_size) {
            data = std::make_unique<uint8_t[]>(in_size);
        }
        auto get_size() const {
            return size;
        }
        auto get_size_bytes() const {
            return size;
        }
        auto get_data() const {
            return data.get();
        }

        operator std::span<uint8_t>() const {
            return {data.get(), size};
        }
        operator std::span<const uint8_t>() const {
            return {data.get(), size};
        }
        operator uint8_t*() const {
            return data.get();
        }
        operator void*() const {
            return data.get();
        }

    private:
        std::unique_ptr<uint8_t[]> data;
        size_t size;
    };

    using heap_ptr = std::shared_ptr<heap>;

    class render_buffer {
    public:
        // 添加预分配大小的常量
        static constexpr size_t DEFAULT_INITIAL_SIZE = 1024;
        static constexpr float GROWTH_FACTOR         = 2.0f;

        render_buffer() : buffer(nullptr), size(0), capacity(0), access_flags(0) {}
        // 禁用拷贝操作
        render_buffer(const render_buffer&)            = delete;
        render_buffer& operator=(const render_buffer&) = delete;

        ~render_buffer();

        void init(const LLGL::BufferDescriptor& in_desc);

        // 预留空间
        void reserve(size_t new_capacity);

        // 收缩到适合大小
        void shrink_to_fit();
        void resize(size_t in_capacity_size, bool in_keep_data = true);
        void clear();

        [[nodiscard]] heap_ptr read(size_t in_offset, size_t in_size) const;
        [[nodiscard]] heap_ptr read() const;

        void push(const void* data, size_t in_data_size);
        template <typename T>
        void push(const std::span<T>& data) {
            push(data.data(), data.size_bytes());
        }
        template<typename T>
        void push(const std::vector<T>& data) {
            push(std::span(data));
        }
        template <class T>
        void push(const T& data) {
            push(&data, sizeof(T));
        }

        void update(const void* data, size_t offset, size_t size);
        template<typename T>
        void update(std::span<const T> data, const size_t offset = 0) {
            update(data.data(), offset, data.size_bytes());
        }
        template<typename T>
        void update(const std::vector<T>& data, const size_t offset = 0) {
            update(std::span(data), offset);
        }
        template<typename T>
        void update(const T& data, const size_t offset = 0) {
            update(&data, offset, sizeof(T));
        }

        void set(const void* in_data, size_t in_size);
        template<typename T>
        void set(std::span<const T> in_data) {
            set(in_data.data(), in_data.size_bytes());
        }
        template<typename T>
        void set(const std::vector<T>& in_data) {
            set(std::span(in_data));
        }
        template<typename T>
        void set(const T& in_data) {
            set(&in_data, sizeof(T));
        }

        // 异步更新缓冲区
        struct update_info {
            const void* data;
            size_t offset;
            size_t size;
        };
        void update_async(const void* data, size_t offset, size_t size);
        void batch_update_async(const std::vector<update_info>& updates);

        [[nodiscard]] size_t get_size() const {
            return size;
        }
        [[nodiscard]] size_t get_capacity() const {
            return capacity;
        }
        [[nodiscard]] bool is_empty() const {
            return size == 0;
        }

        operator LLGL::Buffer*() const {
            return buffer;
        }
        operator bool() const {
            return buffer;
        }
        auto& get_raw() const {
            return *buffer;
        }
    private:
        LLGL::Buffer* buffer;
        size_t size;
        size_t capacity;
        long access_flags;
    };

    template<typename T>
    std::shared_ptr<render_buffer> create_constant_buffer(long in_cpu_access_flags = LLGL::CPUAccessFlags::ReadWrite) {
        LLGL::BufferDescriptor desc;
        desc.size = sizeof(T);
        desc.cpuAccessFlags = in_cpu_access_flags;
        desc.bindFlags = LLGL::BindFlags::ConstantBuffer;
        desc.miscFlags = LLGL::MiscFlags::DynamicUsage;

        auto buffer = std::make_shared<render_buffer>();
        buffer->init(desc);
        return buffer;
    }

    template<typename T = vertex>
    std::shared_ptr<render_buffer> create_vertex_buffer(const size_t in_count, const LLGL::VertexFormat& in_vertex_format = create_vertex_format()) {
        LLGL::BufferDescriptor vertex_buffer_descriptor{};
        vertex_buffer_descriptor.size = sizeof(T) * in_count;
        vertex_buffer_descriptor.bindFlags = LLGL::BindFlags::VertexBuffer;
        vertex_buffer_descriptor.miscFlags = LLGL::MiscFlags::DynamicUsage;
        vertex_buffer_descriptor.cpuAccessFlags = LLGL::CPUAccessFlags::Write;
        vertex_buffer_descriptor.vertexAttribs = in_vertex_format.attributes;
        vertex_buffer_descriptor.stride = sizeof(T);

        auto vertex_buffer = std::make_shared<render_buffer>();
        vertex_buffer->init(vertex_buffer_descriptor);
        return vertex_buffer;
    }

    template<typename T = triangle_index_t>
    std::shared_ptr<render_buffer> create_index_buffer(const size_t in_count) {
        LLGL::BufferDescriptor index_buffer_descriptor{};
        index_buffer_descriptor.size      = sizeof(T) * in_count;
        index_buffer_descriptor.bindFlags = LLGL::BindFlags::IndexBuffer;
        index_buffer_descriptor.miscFlags = LLGL::MiscFlags::DynamicUsage;
        index_buffer_descriptor.cpuAccessFlags = LLGL::CPUAccessFlags::Write;
        index_buffer_descriptor.format = LLGL::Format::R32UInt;

        auto index_buffer = std::make_shared<render_buffer>();
        index_buffer->init(index_buffer_descriptor);
        return index_buffer;
    }
} // namespace mirage
