#include "render_command_pool.h"

#include "mirage.h"
#include <spdlog/spdlog.h>

namespace mirage {

    LLGL::CommandBuffer* render_command_pool::acquire() {
        std::lock_guard lock(mutex);

        // 检查是否有可用的命令缓冲区
        for (auto& buffer : buffers) {
            if (!buffer.in_use) {
                buffer.in_use = true;
                return buffer.cmd_buffer;
            }
        }

        // 如果没有可用的，创建新的
        return create_new_buffer();
    }
    void render_command_pool::release(const LLGL::CommandBuffer* in_cmd_buffer) {
        std::lock_guard lock(mutex);

        for (auto& buffer : buffers) {
            if (buffer.cmd_buffer == in_cmd_buffer) {
                buffer.in_use = false;
                return;
            }
        }
    }
    void render_command_pool::cleanup() {
        std::lock_guard lock(mutex);

        for (auto& buffer : buffers) {
            if (buffer.cmd_buffer) {
                get_renderer()->Release(*buffer.cmd_buffer);
            }
        }
        buffers.clear();
    }
    LLGL::CommandBuffer* render_command_pool::create_new_buffer() {
        if (buffers.size() >= MAX_POOL_SIZE) {
            spdlog::warn("Command pool reached maximum size");
            return nullptr;
        }

        LLGL::CommandBufferDescriptor desc;
        desc.flags = LLGL::CommandBufferFlags::ImmediateSubmit;

        auto* cmd_buffer = get_renderer()->CreateCommandBuffer(desc);
        if (!cmd_buffer) {
            spdlog::error("Failed to create command buffer");
            return nullptr;
        }

        buffers.push_back({cmd_buffer, true});
        return cmd_buffer;
    }
} // namespace mirage
