#pragma once
#include "LLGL/LLGL.h"
#include "misc/lazy_singleton.h"
#include <mutex>

namespace mirage {
    class render_command_pool {
        friend class lazy_singleton_func;
    public:
        // 获取一个可用的命令缓冲区
        LLGL::CommandBuffer* acquire();

        // 释放命令缓冲区回池
        void release(const LLGL::CommandBuffer* in_cmd_buffer);

        // 清理所有资源
        void cleanup();
    private:
        struct buffer_entry {
            LLGL::CommandBuffer* cmd_buffer = nullptr;
            bool in_use                     = false;
        };

        std::vector<buffer_entry> buffers;
        std::mutex mutex;

        static constexpr size_t INITIAL_POOL_SIZE = 4;
        static constexpr size_t MAX_POOL_SIZE     = 16;

        render_command_pool() {
            // 预创建一些命令缓冲区
            for (size_t i = 0; i < INITIAL_POOL_SIZE; ++i) {
                create_new_buffer();
            }
        }

        ~render_command_pool() {
            cleanup();
        }

        LLGL::CommandBuffer* create_new_buffer();
    };

    // RAII 包装器用于自动管理命令缓冲区的获取和释放
    class scoped_command {
    public:
        scoped_command() : cmd_buffer(lazy_singleton<render_command_pool>::get().acquire()) {
            if (cmd_buffer) {
                cmd_buffer->Begin();
            }
        }

        ~scoped_command() {
            if (cmd_buffer) {
                cmd_buffer->End();
                lazy_singleton<render_command_pool>::get().release(cmd_buffer);
            }
        }

        operator LLGL::CommandBuffer*() const { return cmd_buffer; }
        LLGL::CommandBuffer* operator->() const { return cmd_buffer; }

        [[nodiscard]] bool is_valid() const { return cmd_buffer != nullptr; }
    private:
        LLGL::CommandBuffer* cmd_buffer;
    };
} // namespace mirage
