#include "misc/lazy_singleton.h"
#include "window/window.h"
#include "window/window_manager.h"
#include <spdlog/spdlog.h>

namespace mirage {
    class desktop_window_event_listener : public LLGL::Window::EventListener {
    public:
        virtual void OnQuit(LLGL::Window& sender, bool& veto) override {
            auto wrapper = static_cast<window*>(sender.GetUserData());
            wrapper->on_quit(veto);
            lazy_singleton<window_manager>::get().destroy_window(*wrapper);
        }
        virtual void OnKeyDown(LLGL::Window& sender, LLGL::Key keyCode) override {
            auto wrapper = static_cast<window*>(sender.GetUserData());
            wrapper->on_key_down(keyCode);
        }
        virtual void OnKeyUp(LLGL::Window& sender, LLGL::Key keyCode) override {
            auto wrapper = static_cast<window*>(sender.GetUserData());
            wrapper->on_key_up(keyCode);
        }
        virtual void OnDoubleClick(LLGL::Window& sender, LLGL::Key keyCode) override {
            auto wrapper = static_cast<window*>(sender.GetUserData());
            wrapper->on_double_click(keyCode);
        }
        virtual void OnChar(LLGL::Window& sender, wchar_t chr) override {
            auto wrapper = static_cast<window*>(sender.GetUserData());
            wrapper->on_char(chr);
        }
        virtual void OnWheelMotion(LLGL::Window& sender, int motion) override {
            auto wrapper = static_cast<window*>(sender.GetUserData());
            wrapper->on_wheel_motion(motion);
        }
        virtual void OnGlobalMotion(LLGL::Window& sender, const LLGL::Offset2D& motion) override {
            auto wrapper = static_cast<window*>(sender.GetUserData());
            wrapper->on_global_motion(motion);
        }
        virtual void OnResize(LLGL::Window& sender, const LLGL::Extent2D& clientAreaSize) override {
            auto wrapper = static_cast<window*>(sender.GetUserData());
            wrapper->on_resize(clientAreaSize);
        }
        virtual void OnUpdate(LLGL::Window& sender) override {
            auto wrapper = static_cast<window*>(sender.GetUserData());
            wrapper->on_update();
        }
        virtual void OnGetFocus(LLGL::Window& sender) override {
            auto wrapper = static_cast<window*>(sender.GetUserData());
            wrapper->on_get_focus();
        }
        virtual void OnLostFocus(LLGL::Window& sender) override {
            auto wrapper = static_cast<window*>(sender.GetUserData());
            wrapper->on_lost_focus();
        }
    private:

    };

    void window::set_title(const std::u8string& in_title) {
        if (auto window_ptr = static_cast<LLGL::Window*>(surface.get())) {
            window_ptr->SetTitle((char*)in_title.c_str());
        }
    }

    std::u8string window::get_title() const {
        if (const auto window_ptr = static_cast<LLGL::Window*>(surface.get())) {
            return (char8_t*)window_ptr->GetTitle().c_str();
        }
        return {};
    }

    void window::set_fullscreen(bool in_fullscreen) {
        if (auto window_ptr = static_cast<LLGL::Window*>(surface.get())) {
            window_ptr->AdaptForVideoMode(nullptr, &in_fullscreen);
            const auto display = window_ptr->FindResidentDisplay();
            if (display && in_fullscreen) {
                resize(display->GetDisplayMode().resolution);
            }
        }
    }

    bool window::is_fullscreen() const {
        if (const auto window_ptr = static_cast<LLGL::Window*>(surface.get())) {
            if (const auto display = window_ptr->FindResidentDisplay()) {
                const auto& window_client_size = window_ptr->GetSize();
                const auto& display_mode       = display->GetDisplayMode();
                return is_borderless() && window_client_size == display_mode.resolution;
            }
        }
        return false;
    }

    void window::set_resizable(bool in_resizable) {
        if (auto window_ptr = static_cast<LLGL::Window*>(surface.get())) {
            auto desc = window_ptr->GetDesc();
            desc.flags = in_resizable ? desc.flags | LLGL::WindowFlags::Resizable : desc.flags & ~LLGL::WindowFlags::Resizable;
            window_ptr->SetDesc(desc);
        }
    }

    bool window::is_resizable() const {
        if (const auto window_ptr = static_cast<LLGL::Window*>(surface.get())) {
            return window_ptr->GetDesc().flags & LLGL::WindowFlags::Resizable;
        }
        return false;
    }

    void window::set_centered(bool in_centered) {
        if (auto window_ptr = static_cast<LLGL::Window*>(surface.get())) {
            auto desc = window_ptr->GetDesc();
            desc.flags =
                in_centered ? desc.flags | LLGL::WindowFlags::Centered : desc.flags & ~LLGL::WindowFlags::Centered;
            window_ptr->SetDesc(desc);
        }
    }

    void window::set_borderless(bool borderless) {
        if (auto window_ptr = static_cast<LLGL::Window*>(surface.get())) {
            auto desc = window_ptr->GetDesc();
            desc.flags =
                borderless ? desc.flags | LLGL::WindowFlags::Borderless : desc.flags & ~LLGL::WindowFlags::Borderless;
            window_ptr->SetDesc(desc);
        }
    }

    bool window::is_borderless() const {
        if (const auto window_ptr = static_cast<LLGL::Window*>(surface.get())) {
            return window_ptr->GetDesc().flags & LLGL::WindowFlags::Borderless;
        }
        return false;
    }

    void window::show(bool in_show) {
        if (auto window_ptr = static_cast<LLGL::Window*>(surface.get())) {
            window_ptr->Show(in_show);
        }
    }

    void window::move(const LLGL::Offset2D& position) {
        if (auto window_ptr = static_cast<LLGL::Window*>(surface.get())) {
            window_ptr->SetPosition(position);
        }
    }

    void window::move_delta(const LLGL::Offset2D& delta) {
        if (const auto window_ptr = static_cast<LLGL::Window*>(surface.get())) {
            window_ptr->SetPosition(window_ptr->GetPosition() + delta);
        }
    }

    void window::resize(const LLGL::Extent2D& size) {
        if (get_size() == size)
            return;
        if (context) {
            context->resize_swap_chain(size, is_fullscreen() ? LLGL::ResizeBuffersFlags::FullscreenMode : LLGL::ResizeBuffersFlags::WindowedMode);
        } else if (auto window_ptr = static_cast<LLGL::Window*>(surface.get())) {
            if (!window_ptr->AdaptForVideoMode((LLGL::Extent2D*) &size, nullptr)) {
                spdlog::error("无法调整窗口大小");
            }
        }
    }
    LLGL::Extent2D window::get_size(bool in_client_size) const {
        if (const auto window_ptr = static_cast<LLGL::Window*>(surface.get())) {
            return window_ptr->GetSize(in_client_size);
        }
        return {};
    }

    void window::on_resize(const LLGL::Extent2D& in_client_area_size) {
        if (context) {
            context->resize_swap_chain(in_client_area_size, 0);
        }
    }

    void window::init_event_listener() {
        if (auto window_ptr = static_cast<LLGL::Window*>(surface.get())) {
            window_ptr->SetUserData(this);
            window_ptr->AddEventListener(std::make_shared<desktop_window_event_listener>());
        }
    }
}
