#include "window_manager.h"

#include <spdlog/spdlog.h>

namespace mirage {
    void window_manager::update(const duration_type& in_delta_time) {
        LLGL::Surface::ProcessEvents();
        render_thread_func();
        to_destroy.clear();
        should_exit_flag = windows.empty() && to_destroy.empty();
    }

    bool window_manager::init(const init_info& in_info) {
        auto main_window = create_window(in_info);
        auto window_ptr  = main_window.lock();
        if (!window_ptr) {
            return false;
        }

        if (!window_ptr->create_context(in_info)) {
            return false;
        }
        last_time     = std::chrono::high_resolution_clock::now();
        return true;
    }

    bool window_manager::destroy() {
        should_exit_flag = true;
        windows.clear();
        to_destroy.clear();
        return true;
    }

    std::weak_ptr<window> window_manager::create_window(const LLGL::WindowDescriptor& desc) {
        auto window_ptr = std::make_shared<window>(desc);

        windows.push_back(window_ptr);
        if (!window_ptr) {
            spdlog::error("无法创建窗口");
            return {};
        }
        return window_ptr;
    }

    void window_manager::destroy_window(const window& in_window) {
        for (auto it = windows.begin(); it != windows.end(); ++it) {
            if (it->get() == &in_window) {
                to_destroy.push_back(*it);
                windows.erase(it);
                break;
            }
        }
    }

    std::weak_ptr<window> window_manager::get_main_window() {
        if (windows.empty()) {
            return {};
        }
        return windows.front();
    }

    void window_manager::render_thread_func() {
        const auto& current_time = std::chrono::high_resolution_clock::now();
        const auto in_delta_time = current_time - last_time;
        last_time                = current_time;

        for (const auto& w : windows) {
            w->render_update(in_delta_time);
        }
    }
} // namespace mirage
